
#ifndef PLS_WORK_STEALING_DEQUE_H_
#define PLS_WORK_STEALING_DEQUE_H_

#include <atomic>
#include <mutex>
#include <pls/internal/scheduling/thread_state.h>

#include "pls/internal/base/system_details.h"
#include "pls/internal/base/spin_lock.h"
#include "pls/internal/base/error_handling.h"

#include "aligned_stack.h"

namespace pls {
namespace internal {
namespace data_structures {

using cas_integer = base::system_details::cas_integer;
using pointer_t = base::system_details::pointer_t;
static cas_integer get_jump_wish(cas_integer n) {
  return (n & base::system_details::FIRST_HALF_CAS_INTEGER) >> ((sizeof(cas_integer) / 2) * 8);
}
static cas_integer get_offset(cas_integer n) {
  return n & base::system_details::SECOND_HALF_CAS_INTEGER;
}
static cas_integer set_jump_wish(cas_integer n, cas_integer new_value) {
  return (new_value << ((sizeof(cas_integer) / 2) * 8)) | (n & base::system_details::SECOND_HALF_CAS_INTEGER);
}
static cas_integer set_offset(cas_integer n, cas_integer new_value) {
  return new_value | (n & base::system_details::FIRST_HALF_CAS_INTEGER);
}

class work_stealing_deque_item {
  // Pointer to the actual data
  pointer_t data_;
  // Index (relative to stack base) to the next and previous element
  cas_integer next_item_;
  cas_integer previous_item_;

 public:
  work_stealing_deque_item() : data_{0}, next_item_{0}, previous_item_{0} {}

  template<typename Item>
  Item *data() {
    return reinterpret_cast<Item *>(data_);
  }

  template<typename Item>
  void set_data(Item *data) {
    data_ = reinterpret_cast<pointer_t >(data);
  }

  cas_integer next_item() {
    return next_item_;
  }
  void set_next_item(cas_integer next_item) {
    next_item_ = next_item;
  }
  cas_integer previous_item() {
    return previous_item_;
  }
  void set_previous_item(cas_integer previous_item) {
    previous_item_ = previous_item;
  }
};
static_assert(sizeof(work_stealing_deque_item) < base::system_details::CACHE_LINE_SIZE,
              "Work stealing deque relies on memory layout and requires cache lines to be longer than one 'work_stealing_deque_item' instance!");

template<typename Item>
class work_stealing_deque {
  // Deque 'takes over' stack and handles memory management while in use.
  // At any point in time the deque can stop using more memory and the stack can be used by other entities.
  aligned_stack *stack_;
  pointer_t base_pointer_;

  std::atomic<cas_integer> head_;
  std::atomic<cas_integer> tail_;
  cas_integer previous_tail_;

  base::spin_lock lock_{}; // TODO: Remove after debugging


 public:
  using state = aligned_stack::state;

  explicit work_stealing_deque(aligned_stack *stack) : stack_{stack}, head_{0}, tail_{0}, previous_tail_{0} {
    reset_base_pointer();
  }
  work_stealing_deque(const work_stealing_deque &other) : stack_{other.stack_},
                                                          base_pointer_{other.base_pointer_},
                                                          head_{other.head_.load()},
                                                          tail_{other.tail_.load()},
                                                          previous_tail_{other.previous_tail_} {}

  void reset_base_pointer() {
    base_pointer_ = reinterpret_cast<pointer_t >(stack_->save_state()); // Keep the base of our region in the stack
  }

  work_stealing_deque_item *item_at(cas_integer position) {
    return reinterpret_cast<work_stealing_deque_item *>(base_pointer_
        + (base::system_details::CACHE_LINE_SIZE * position));
  }

  cas_integer current_stack_offset() {
    return (stack_->save_state() - base_pointer_) / base::system_details::CACHE_LINE_SIZE;
  }

  template<typename T>
  std::pair<work_stealing_deque_item, T> *allocate_item(const T &new_item) {
    // 'Union' type to push both on stack
    using pair_t = std::pair<work_stealing_deque_item, T>;
    // Allocate space on stack
    auto new_pair = reinterpret_cast<pair_t *>(stack_->push<pair_t>());
    // Initialize memory on stack
    new((void *) &(new_pair->first)) work_stealing_deque_item();
    new((void *) &(new_pair->second)) T(new_item);

    return new_pair;
  }

  template<typename T>
  Item *push_tail(const T &new_item) {
//    std::lock_guard<base::spin_lock> lock{lock_};
    cas_integer local_tail = tail_;
    cas_integer local_head = head_;
//    PLS_ASSERT((local_tail >= get_offset(local_head)), "Tail MUST be in front of head!")
    auto new_pair = allocate_item(new_item);

    // Prepare current tail to point to correct next items
    auto tail_deque_item = item_at(local_tail);
    tail_deque_item->set_data(&(new_pair->second));
    tail_deque_item->set_next_item(current_stack_offset());
    tail_deque_item->set_previous_item(previous_tail_);
    previous_tail_ = local_tail;

    // Linearization point, item appears after this write
    cas_integer new_tail = current_stack_offset();
    tail_ = new_tail;
//    {
//      std::lock_guard<base::spin_lock> lock{lock_};
//      std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                << "Pushed Tail " << local_tail << "->" << new_tail << std::endl;
//    }
  }

  Item *pop_tail() {
//    std::lock_guard<base::spin_lock> lock{lock_};
    cas_integer local_tail = tail_;
    cas_integer local_head = head_;

    if (local_tail <= get_offset(local_head)) {
      return nullptr; // EMPTY
    }

    work_stealing_deque_item *previous_tail_item = item_at(previous_tail_);
    cas_integer new_tail = previous_tail_;
    previous_tail_ = previous_tail_item->previous_item();

    // Publish our wish to set the tail back
    tail_ = new_tail;
    // Get the state of local head AFTER we published our wish
    local_head = head_; // Linearization point, outside knows list is empty

    if (get_offset(local_head) < new_tail) {
//      {
//        std::lock_guard<base::spin_lock> lock{lock_};
//        std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                  << "Poped Tail (distance) " << local_tail << "->" << new_tail << std::endl;
//      }
      return previous_tail_item->data<Item>(); // Enough distance, return item
    }

    cas_integer new_head = set_jump_wish(new_tail, 999999);
    if (get_offset(local_head) == new_tail) {
      // Try competing with consumers...
      if (head_.compare_exchange_strong(local_head, new_head)) {
//        {
//          std::lock_guard<base::spin_lock> lock{lock_};
//          std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                    << "Poped Tail (won competition 1) " << local_tail << "->" << new_tail << std::endl;
//        }
        return previous_tail_item->data<Item>(); // We won the competition, linearization on whom got the item
      }
      // Cosumer either registered jump wish or has gotten the item.
      // Local_Head has the new value of the head, see if the other thread got to advance it
      // and if not (only jump wish) try to win the competition.
      if (get_offset(local_head) == new_tail && head_.compare_exchange_strong(local_head, new_head)) {
//        {
//          std::lock_guard<base::spin_lock> lock{lock_};
//          std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                    << "Poped Tail (won competition 2) " << local_tail << "->" << new_tail << std::endl;
//        }
        return previous_tail_item->data<Item>(); // We won the competition, linearization on whom got the item
      }
    }
//    {
//      std::lock_guard<base::spin_lock> lock{lock_};
//      std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                << "FAILED to pop tail (lost competition) " << get_offset(local_head) << "; " << local_tail << "->"
//                << new_tail << std::endl;
//    }

    // Some other thread either won the competition or it already set the head further than we are
    // before we even tried to compete with it.
    // Reset the queue into an empty state => head_ = tail_
    // We can not set it to 0, as the memory is still in use.
    tail_ = get_offset(local_head); // Set tail to match the head value the other thread won the battle of

    return nullptr;
  }

  Item *pop_head() {
//    std::lock_guard<base::spin_lock> lock{lock_};
    cas_integer local_tail = tail_;
    cas_integer local_head = head_;
    cas_integer local_head_offset = get_offset(local_head);

    if (local_head_offset >= local_tail) {
      return nullptr; // EMPTY
    }
    work_stealing_deque_item *head_deque_item = item_at(local_head_offset);
    cas_integer next_item_offset = head_deque_item->next_item();
    Item *head_data_item = head_deque_item->data<Item>();

    cas_integer jump_wish_head = set_jump_wish(local_head_offset, head_deque_item->next_item());
    if (!head_.compare_exchange_strong(local_head, jump_wish_head)) {
//      {
//        std::lock_guard<base::spin_lock> lock{lock_};
//        std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                  << "Failed to pop head (first cas) " << local_head_offset << "->" << next_item_offset << std::endl;
//      }
      return nullptr; // Someone interrupted us
    }

    local_tail = tail_;
    if (local_head_offset >= local_tail) {
//      std::cout << "Failed to pop head (second tail test) " << get_offset(local_head) << std::endl;
      return nullptr; // EMPTY, tail was removed while we registered our jump wish
    }

    cas_integer new_head = next_item_offset;
    if (!head_.compare_exchange_strong(jump_wish_head, new_head)) {
//      {
//        std::lock_guard<base::spin_lock> lock{lock_};
//        std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                  << "Failed to pop head (second cas) " << local_head_offset << "->" << next_item_offset << std::endl;
//      }
      return nullptr; // we lost the 'fight' on the item...
    }

//    {
//      std::lock_guard<base::spin_lock> lock{lock_};
//      std::cout << base::this_thread::state<scheduling::thread_state>()->id_ << " - "
//                << "Popped Head " << local_head_offset << "->" << next_item_offset << std::endl;
//    }
    return head_deque_item->data<Item>(); // We won the 'fight' on the item, it is now save to access it!
  }

  void release_memory_until(state state) {
//    std::lock_guard<base::spin_lock> lock{lock_};
    cas_integer
        item_offset = (state - base_pointer_) / base::system_details::CACHE_LINE_SIZE;

    cas_integer local_head = head_;
    cas_integer local_tail = tail_;

//    if (local_tail != item_offset) {
//      std::cout << "...";
//    } else {
//      std::cout << "...";
//    }

    stack_->reset_state(state);

    if (item_offset < local_tail) {
      tail_ = item_offset;
      if (get_offset(local_head) >= local_tail) {
        head_ = item_offset;
      }
    }

//    std::cout << "Release Memory " << item_offset << std::endl;
  }

  void release_memory_until(Item *item) {
    release_memory_until(reinterpret_cast<pointer_t >(item));
  }

  state save_state() {
    return stack_->save_state();
  }

  // PUSH item onto stack (allocate + insert into stack) - CHECK
  // POP item from bottom of stack (remove from stack, memory still used) - CHECK
  // POP item from top of stack (remove from stack, memory still used) - CHECK

  // RELEASE memory from all items allocated after this one (including this one) - CHECK
  // -> Tell the data structure that it is safe to reuse the stack space
  // Note: Item that is released must not be part of the queue at this point (it is already removed!)
};

}
}
}

#endif //PLS_WORK_STEALING_DEQUE_H_
